import unittest
import time
import threading

from pyrediseasyio.io_group import IOGroup
from pyrediseasyio.reader_writer import ReaderWriter
from assertpy import assert_that




class PubSubTests(unittest.TestCase):

    def _publish_many(self, count: int, wait: float = 0):
        if wait > 0:
            time.sleep(wait)
        for i in range(0, count):
            self.pubsub1.publish(f"Message {i}")

    def setUp(self):
        self.pubsub1 = ReaderWriter(channel="Channel1")
        self.pubsub2 = ReaderWriter(channel="Channel1")
        self.pubsub1.flush_db()
        self.pubsub2.flush_db()

    def test_get_single_message(self):
        pubsub1, pubsub2 = self.pubsub1, self.pubsub2
        r = pubsub2.get_next_message()
        assert_that(r).is_none()
        data = "This is a test"
        pubsub1.publish(data)
        c, d = pubsub2.get_next_message()
        assert_that(c).is_equal_to("Channel1")
        assert_that(d).is_equal_to(data)

    def test_get_many_messages(self):
        count = 2
        pubsub1, pubsub2 = self.pubsub1, self.pubsub2
        self._publish_many(count)
        time.sleep(0.2)
        messages = pubsub2.get_messages()
        assert_that(len(messages)).is_equal_to(count)
        messages = pubsub2.get_messages()
        assert_that(messages).is_empty()

    def test_listen_for_messages(self):
        pubsub1, pubsub2 = self.pubsub1, self.pubsub2
        threading.Thread(target=self._publish_many, args=(20, 0.1)).start()
        messages = []
        for i, m in enumerate(pubsub2.listen()):
            if i >= 10:
                break
            messages.append(m[1])
        assert_that(len(messages)).is_equal_to(10)




from pyrediseasyio.reader_writer import ReaderWriter
from pyrediseasyio.single_io import SingleIO
import json


class IOGroup(ReaderWriter):

    def __init__(self, host='localhost', port=6379, db=0,
                 set_defaults_on_startup: bool = False,
                 delete_keys_on_startup: bool = False, **kwargs):

        super().__init__(host=host, port=port, db=db, **kwargs)
        member_names = [d for d in dir(self) if not d.startswith('_')]
        self.members = []
        for name in member_names:
            try:
                attr = getattr(self, name)
                if not isinstance(attr, SingleIO):
                    continue
                attr._reader_writer = self
                self.members.append(name)
                if delete_keys_on_startup:
                    self.delete_key(attr.addr)
                if set_defaults_on_startup:
                    attr.write(attr.default)
            except AttributeError:
                pass

    def __setattr__(self, key, value):
        if hasattr(self, 'members') and key in self.members:
            attr = getattr(self, key)
            attr.write(value)
            return
        return super().__setattr__(key, value)

    def dump(self, key: str) -> str:
        """
        Creates a json string containing a list of dict(name/addr/value) of all members and stores it in-memory
        using the given key. The value can later be retrieved by a call to the 'read' method.
        :param key: A unique key for storing the dumped value
        :return: Returns the json string that is stored - for convenience sake.
        """
        s = self.dumps()
        self.write(key, s)
        return s

    def dumps(self) -> str:
        """
        Returns a json string containing a list of dict(name/addr/value) of all members.
        """
        members = []
        for m in self.members:
            attr = getattr(self, m)
            members.append(dict(name=attr.name, addr=attr.addr, value=attr.value))
        return json.dumps(members)







from pyrediseasyio.abstract_reader_writer import AbstractReaderWriter
import threading
from str2bool import str2bool
import json

lock = threading.Lock()


class SingleIO:
    def __init__(self, name: str, addr: str, default: object, units: str = None, reader: AbstractReaderWriter = None):
        self.name = name
        self.addr = addr
        self._reader_writer = reader
        self.default = default
        self.units = units

    def __and__(self, other):
        if hasattr(other, 'value'):
            return self.value and other.value
        return self.value and other

    def __add__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value + other

    def __sub__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value - other

    def __mul__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value * other

    def __eq__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value == other

    def __floordiv__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value // other

    def __iadd__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        val = self.value
        val += other
        self.write(val)

    def __idiv__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        val = self.value
        val /= other
        self.write(val)

    def __ifloordiv__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        val = self.value
        val //= other
        self.write(val)

    def __imul__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        val = self.value
        val *= other
        self.write(val)

    def __invert__(self):
        return ~self.value

    def __ipow__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        val = self.value
        val **= other
        self.write(val)

    def __isub__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        val = self.value
        val -= other
        self.write(val)

    def __ne__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value != other

    def __or__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value or other

    def __pos__(self):
        return self.value

    def __pow__(self, power, modulo=None):
        if hasattr(power, 'value'):
            power = power.value
        return self.value ** power

    def __neg__(self):
        return -self.value

    def __set__(self, obj, value):
        self.write(value)

    def __str__(self):
        return f'[{type(self).__name__}] {self.name} = {self.value} {self.units}'

    def __truediv__(self, other):
        if hasattr(other, 'value'):
            other = other.value
        return self.value / other

    @property
    def value(self):
        return self.read()

    @staticmethod
    def _convert_type(value):
        return value

    def publish(self, value, channel: str = None):
        value = self._convert_type(value)
        data = json.dumps({self.addr: value})
        self._reader_writer.publish(data, channel)

    def read(self):
        if self._reader_writer is None:
            return None
        with lock:
            val = self._reader_writer.read(self.addr)
            if val is None:
                val = self.default
            val = self._convert_type(val)
            return val

    def write(self, value):
        if self._reader_writer is None:
            return
        with lock:
            value = self._convert_type(value)
            self._reader_writer.write(self.addr, value)


class BooleanIO(SingleIO):
    def __init__(self, name: str, addr: str, default: bool = False, units: str = None, reader: AbstractReaderWriter = None):
        super().__init__(name, addr, default, units, reader)

    @staticmethod
    def _convert_type(value):
        if isinstance(value, str):
            return str2bool(value)
        return bool(value)

    def __bool__(self):
        return self.value
    
    @property
    def value(self) -> bool:
        return super().value


class IntIO(SingleIO):
    def __init__(self, name: str, addr: str, default: int = 0, units: str = None, reader: AbstractReaderWriter = None):
        super().__init__(name, addr, default, units, reader)

    @staticmethod
    def _convert_type(value):
        return int(value)


class FloatIO(SingleIO):
    def __init__(self, name: str, addr: str, default: float = 0, units: str = None, reader: AbstractReaderWriter = None):
        super().__init__(name, addr, default, units, reader)

    @staticmethod
    def _convert_type(value):
        return float(value)


class StringIO(SingleIO):
    def __init__(self, name: str, addr: str, default: str = '', units: str = None, reader: AbstractReaderWriter = None):
        super().__init__(name, addr, default, units, reader)

    @staticmethod
    def _convert_type(value):
        return str(value)


